# -*- coding: utf-8 -*-

# Copyright (c) 2008 - 2009 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""
Module implementing the TurboGears project plugin.
"""

import sys
import os
import glob

from PyQt4.QtCore import QObject, SIGNAL, QTranslator, QString

from KdeQt.KQApplication import e4App

from Project.ProjectBrowser import SourcesBrowserFlag, FormsBrowserFlag, \
    ResourcesBrowserFlag, TranslationsBrowserFlag, InterfacesBrowserFlag, \
    OthersBrowserFlag, AllBrowsersFlag

import Preferences

from ProjectTurboGears.Project import Project

# Start-of-Header
name = "TurboGears Project Plugin"
author = "Detlev Offenbach <detlev@die-offenbachs.de>"
autoactivate = True
deactivateable = True
version = "1.0.5"
className = "ProjectTurboGearsPlugin"
packageName = "ProjectTurboGears"
shortDescription = "Project support for TurboGears projects."
longDescription = """This plugin implements project support for TurboGears projects."""
needsRestart = False
# End-of-Header

error = QString("")

def apiFiles(language):
    """
    Module function to return the API files made available by this plugin.
    
    @return list of API filenames (list of string)
    """
    if language == "Python":
        apisDir = \
            os.path.join(os.path.dirname(__file__), "ProjectTurboGears", "APIs")
        apis = glob.glob(os.path.join(apisDir, '*.api'))
    else:
        apis = []
    return apis

def prepareUninstall():
    """
    Module function to prepare for an uninstallation.
    """
    turbogears = ProjectTurboGearsPlugin(None)
    turbogears.prepareUninstall()
    
class ProjectTurboGearsPlugin(QObject):
    """
    Class implementing the TurboGears project plugin.
    """
    def __init__(self, ui):
        """
        Constructor
        
        @param ui reference to the user interface object (UI.UserInterface)
        """
        QObject.__init__(self, ui)
        self.__ui = ui
        self.__initialize()
        
        self.__translator = None
        self.__loadTranslator()
    
    def __initialize(self):
        """
        Private slot to (re)initialize the plugin.
        """
        self.__object = None
        
        self.__mainAct = None
        self.__mainMenu = None
        
        self.__e4project = e4App().getObject("Project")
    
    def __checkVersions(self):
        """
        Private function to check that the eric4 version is ok.
        
        @return flag indicating version is ok (boolean)
        """
        global error
        
        if self.__ui.versionIsNewer('4.1.99', '20080229'):
            error = QString("")
        else:
            error = self.trUtf8("eric4 version is too old, %1, %2 or newer needed.")\
                .arg("4.2.0").arg("20080229")
            return False
        
        return True
    
    def activate(self):
        """
        Public method to activate this plugin.
        
        @return tuple of None and activation status (boolean)
        """
        if not self.__checkVersions():
            return None, False
        
        self.__object = Project(self, self.__ui)
##        self.__object.initActions()
        e4App().registerPluginObject("ProjectTurboGears", self.__object)
        
##        self.__mainMenu = self.__object.initMenu()
        
        self.__e4project.registerProjectType("TurboGears", self.trUtf8("TurboGears"), 
            self.fileTypesCallback)
        Preferences.setProjectBrowserFlagsDefault("TurboGears", 
            SourcesBrowserFlag | \
            FormsBrowserFlag | \
            TranslationsBrowserFlag | \
            OthersBrowserFlag,
        )
        
        if self.__e4project.isOpen():
##            self.__projectOpened()
            self.__object.projectOpenedHooks()
        
##        self.connect(e4App().getObject("Project"), SIGNAL('projectOpened'),
##                     self.__projectOpened)
##        self.connect(e4App().getObject("Project"), SIGNAL('projectClosed'),
##                     self.__projectClosed)
##        self.connect(e4App().getObject("Project"), SIGNAL('newProject'),
##                     self.__projectOpened)
        
        self.connect(e4App().getObject("Project"), SIGNAL('projectOpenedHooks'),
                     self.__object.projectOpenedHooks)
        self.connect(e4App().getObject("Project"), SIGNAL('projectClosedHooks'),
                     self.__object.projectClosedHooks)
        self.connect(e4App().getObject("Project"), SIGNAL('newProjectHooks'),
                     self.__object.projectOpenedHooks)
        
        return None, True
    
    def deactivate(self):
        """
        Public method to deactivate this plugin.
        """
        e4App().unregisterPluginObject("ProjectTurboGears")
        
##        self.disconnect(e4App().getObject("Project"), SIGNAL('projectOpened'),
##                        self.__projectOpened)
##        self.disconnect(e4App().getObject("Project"), SIGNAL('projectClosed'),
##                        self.__projectClosed)
##        self.disconnect(e4App().getObject("Project"), SIGNAL('newProject'),
##                        self.__projectOpened)
        
        self.disconnect(e4App().getObject("Project"), SIGNAL('projectOpenedHooks'),
                        self.__object.projectOpenedHooks)
        self.disconnect(e4App().getObject("Project"), SIGNAL('projectClosedHooks'),
                        self.__object.projectClosedHooks)
        self.disconnect(e4App().getObject("Project"), SIGNAL('newProjectHooks'),
                        self.__object.projectOpenedHooks)
        
        self.__e4project.unregisterProjectType("TurboGears")
        
        self.__object.projectClosedHooks()
##        self.__projectClosed()
        
        self.__initialize()
    
    def __loadTranslator(self):
        """
        Private method to load the translation file.
        """
        if self.__ui is not None:
            loc = self.__ui.getLocale()
            if loc and loc != "C":
                locale_dir = \
                    os.path.join(os.path.dirname(__file__), "ProjectTurboGears", "i18n")
                translation = "turbogears_%s" % loc
                translator = QTranslator(None)
                loaded = translator.load(translation, locale_dir)
                if loaded:
                    self.__translator = translator
                    e4App().installTranslator(self.__translator)
                else:
                    print "Warning: translation file '%s' could not be loaded." % \
                        translation
                    print "Using default."
    
    def __projectOpened(self):
        """
        Private slot to handle the projectOpened signal.
        """
        if self.__e4project.getProjectType() == "TurboGears":
            projectAct = self.__ui.getMenuBarAction("project")
            actions = self.__ui.menuBar().actions()
            insertAct = actions[actions.index(projectAct) + 1]
            self.__mainAct = self.__ui.menuBar().insertMenu(insertAct, self.__mainMenu)
    
    def __projectClosed(self):
        """
        Private slot to handle the projectClosed signal.
        """
        if self.__mainAct is not None:
            self.__ui.menuBar().removeAction(self.__mainAct)
            self.__mainAct = None
    
    def fileTypesCallback(self):
        """
        Public method get the filetype associations of the TurboGears project type.
        
        @return dictionary with file type associations
        """
        if self.__e4project.getProjectType() == "TurboGears":
            fileTypes = {
                "*.html" : "FORMS", 
                "*.htm"  : "FORMS", 
                "*.kid"  : "FORMS", 
                "*.js"   : "FORMS", 
            }
        else:
            fileTypes = {}
        return fileTypes
    
    def prepareUninstall(self):
        """
        Public method to prepare for an uninstallation.
        """
        Preferences.removeProjectBrowserFlags("TurboGears")
