# -*- coding: utf-8 -*-

# Copyright (c) 2007 - 2010 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""
Module implementing the PyLint plugin.
"""

import os
import sys
import copy

from PyQt4.QtCore import QObject, SIGNAL, QString, QTranslator
from PyQt4.QtGui import QDialog, QApplication

from KdeQt.KQApplication import e4App

from E4Gui.E4Action import E4Action

from PyLint.PyLintConfigDialog import PyLintConfigDialog
from PyLint.PyLintExecDialog import PyLintExecDialog

import Utilities

# Start-of-Header
name = "PyLint Plugin"
author = "Detlev Offenbach <detlev@die-offenbachs.de>"
autoactivate = True
deactivateable = True
version = "4.2.1"
className = "PyLintPlugin"
packageName = "PyLint"
shortDescription = "Show the PyLint dialogs."
longDescription = """This plugin implements the PyLint dialogs.""" \
 """ PyLint is used to check Python source files according to various rules."""
needsRestart = False
# End-of-Header

error = QString("")

def exeDisplayData():
    """
    Public method to support the display of some executable info.
    
    @return dictionary containing the data to query the presence of
        the executable
    """
    exe = 'pylint'
    if sys.platform == "win32":
        ex_ = os.path.join(sys.exec_prefix, "Scripts", exe + '.bat')
        if os.path.exists(ex_):
            exe = ex_
        else:
            exe = exe + '.bat'
    
    data = {
        "programEntry"      : True, 
        "header"            : QApplication.translate("PyLintPlugin",
                                "Checkers - Pylint"), 
        "exe"               : exe, 
        "versionCommand"    : '--version', 
        "versionStartsWith" : 'pylint', 
        "versionPosition"   : -1, 
        "version"           : "", 
        "versionCleanup"    : (0, -1), 
    }
    
    return data

class PyLintPlugin(QObject):
    """
    Class implementing the PyLint plugin.
    """
    def __init__(self, ui):
        """
        Constructor
        
        @param ui reference to the user interface object (UI.UserInterface)
        """
        QObject.__init__(self, ui)
        self.__ui = ui
        self.__initialize()
        
        self.__translator = None
        self.__loadTranslator()
        
    def __initialize(self):
        """
        Private slot to (re)initialize the plugin.
        """
        self.__projectAct = None
        self.__projectShowAct = None
        self.__pylintPDialog = None
        
        self.__projectBrowserAct = None
        self.__projectBrowserShowAct = None
        self.__projectBrowserMenu = None
        self.__pylintPsbDialog = None
        
        self.__editors = []
        self.__editorAct = None
        self.__editorPylintDialog = None
        self.__editorParms = None

    def activate(self):
        """
        Public method to activate this plugin.
        
        @return tuple of None and activation status (boolean)
        """
        global error
        
        # pylint is only activated if it is available
        exe = 'pylint'
        if sys.platform == "win32":
            ex_ = os.path.join(sys.exec_prefix, "Scripts", exe + '.bat')
            if os.path.exists(ex_):
                exe = ex_
            else:
                exe = exe + '.bat'
        if Utilities.isinpath(exe):
            try:
                from pylint.__pkginfo__ import numversion
                if numversion < (0, 12, 0):
                    error = self.trUtf8("PyLint version < 0.12.0.")
                    return None, False
            except ImportError:
                error = self.trUtf8("Cannot determine Pylint version.")
                return None, False
        else:
            error = self.trUtf8("PyLint executable '%1' not found.").arg(exe)
            return None, False
        
        menu = e4App().getObject("Project").getMenu("Checks")
        if menu:
            self.__projectAct = E4Action(self.trUtf8('Run PyLint'),
                    self.trUtf8('Run &PyLint...'), 0, 0,
                    self, 'project_check_pylint')
            self.__projectAct.setStatusTip(\
                self.trUtf8('Check project, packages or modules with pylint.'))
            self.__projectAct.setWhatsThis(self.trUtf8(
                """<b>Run PyLint...</b>"""
                """<p>This checks the project, packages or modules using pylint.</p>"""
            ))
            self.connect(self.__projectAct, SIGNAL('triggered()'), self.__projectPylint)
            e4App().getObject("Project").addE4Actions([self.__projectAct])
            menu.addAction(self.__projectAct)
            
            self.__projectShowAct = E4Action(self.trUtf8('Show PyLint Dialog'),
                    self.trUtf8('Show Py&Lint Dialog...'), 0, 0,
                    self,'project_check_pylintshow')
            self.__projectShowAct.setStatusTip(\
                self.trUtf8('Show the PyLint dialog with the results of the last run.'))
            self.__projectShowAct.setWhatsThis(self.trUtf8(
                """<b>Show PyLint Dialog...</b>"""
                """<p>This shows the PyLint dialog with the results"""
                """ of the last run.</p>"""
            ))
            self.connect(self.__projectShowAct, SIGNAL('triggered()'), 
                         self.__projectPylintShow)
            e4App().getObject("Project").addE4Actions([self.__projectShowAct])
            menu.addAction(self.__projectShowAct)
        
        self.__editorAct = E4Action(self.trUtf8('Run PyLint'),
                self.trUtf8('Run &PyLint...'), 0, 0,
                self, "")
        self.__editorAct.setWhatsThis(self.trUtf8(
                """<b>Run PyLint...</b>"""
                """<p>This checks the loaded module using pylint.</p>"""
        ))
        self.connect(self.__editorAct, SIGNAL('triggered()'), self.__editorPylint)
        
        self.connect(e4App().getObject("Project"), SIGNAL("showMenu"), 
                     self.__projectShowMenu)
        self.connect(e4App().getObject("ProjectBrowser").getProjectBrowser("sources"), 
                     SIGNAL("showMenu"), self.__projectBrowserShowMenu)
        self.connect(e4App().getObject("ViewManager"), SIGNAL("editorOpenedEd"), 
                     self.__editorOpened)
        self.connect(e4App().getObject("ViewManager"), SIGNAL("editorClosedEd"), 
                     self.__editorClosed)
        
        for editor in e4App().getObject("ViewManager").getOpenEditors():
            self.__editorOpened(editor)
        
        error = QString("")
        return None, True

    def deactivate(self):
        """
        Public method to deactivate this plugin.
        """
        self.disconnect(e4App().getObject("Project"), SIGNAL("showMenu"), 
                        self.__projectShowMenu)
        self.disconnect(e4App().getObject("ProjectBrowser").getProjectBrowser("sources"), 
                        SIGNAL("showMenu"), self.__projectBrowserShowMenu)
        self.disconnect(e4App().getObject("ViewManager"), SIGNAL("editorOpenedEd"), 
                        self.__editorOpened)
        self.disconnect(e4App().getObject("ViewManager"), SIGNAL("editorClosedEd"), 
                        self.__editorClosed)
        
        menu = e4App().getObject("Project").getMenu("Checks")
        if menu:
            if self.__projectAct:
                menu.removeAction(self.__projectAct)
                e4App().getObject("Project").removeE4Actions([self.__projectAct])
            if self.__projectShowAct:
                menu.removeAction(self.__projectShowAct)
                e4App().getObject("Project").removeE4Actions([self.__projectShowAct])
        
        if self.__projectBrowserMenu:
            if self.__projectBrowserAct:
                self.__projectBrowserMenu.removeAction(self.__projectBrowserAct)
            if self.__projectBrowserShowAct:
                self.__projectBrowserMenu.removeAction(self.__projectBrowserShowAct)
        
        for editor in self.__editors:
            self.disconnect(editor, SIGNAL("showMenu"), self.__editorShowMenu)
            menu = editor.getMenu("Checks")
            if menu is not None:
                menu.removeAction(self.__editorAct)
        
        self.__initialize()
    
    def __loadTranslator(self):
        """
        Private method to load the translation file.
        """
        if self.__ui is not None:
            loc = self.__ui.getLocale()
            if loc and loc != "C":
                locale_dir = \
                    os.path.join(os.path.dirname(__file__), "PyLint", "i18n")
                translation = "pylint_%s" % loc
                translator = QTranslator(None)
                loaded = translator.load(translation, locale_dir)
                if loaded:
                    self.__translator = translator
                    e4App().installTranslator(self.__translator)
                else:
                    print "Warning: translation file '%s' could not be loaded." % \
                        translation
                    print "Using default."
    
    def __projectShowMenu(self, menuName, menu):
        """
        Private slot called, when the the project menu or a submenu is 
        about to be shown.
        
        @param menuName name of the menu to be shown (string)
        @param menu reference to the menu (QMenu)
        """
        if menuName == "Checks":
            if self.__projectAct is not None:
                self.__projectAct.setEnabled(\
                    e4App().getObject("Project").getProjectLanguage() == "Python")
            if self.__projectShowAct is not None:
                self.__projectShowAct.setEnabled(\
                    e4App().getObject("Project").getProjectLanguage() == "Python")
            self.__projectShowAct.setEnabled(self.__pylintPDialog is not None)
    
    def __projectBrowserShowMenu(self, menuName, menu):
        """
        Private slot called, when the the project browser menu or a submenu is 
        about to be shown.
        
        @param menuName name of the menu to be shown (string)
        @param menu reference to the menu (QMenu)
        """
        if menuName == "Checks" and \
           e4App().getObject("Project").getProjectLanguage() == "Python":
            self.__projectBrowserMenu = menu
            if self.__projectBrowserAct is None:
                self.__projectBrowserAct = E4Action(self.trUtf8('Run PyLint'),
                        self.trUtf8('Run &PyLint...'), 0, 0,
                        self, '')
                self.__projectBrowserAct.setWhatsThis(self.trUtf8(
                    """<b>Run PyLint...</b>"""
                    """<p>This checks the project, packages or modules"""
                    """ using pylint.</p>"""
                ))
                self.connect(self.__projectBrowserAct, SIGNAL('triggered()'), 
                             self.__projectBrowserPylint)
            
            if self.__projectBrowserShowAct is None:
                self.__projectBrowserShowAct = \
                        E4Action(self.trUtf8('Show PyLint Dialog'),
                        self.trUtf8('Show Py&Lint Dialog...'), 0, 0,
                        self,'')
                self.__projectBrowserShowAct.setWhatsThis(self.trUtf8(
                    """<b>Show PyLint Dialog...</b>"""
                    """<p>This shows the PyLint dialog with the results"""
                    """ of the last run.</p>"""
                ))
                self.connect(self.__projectBrowserShowAct, SIGNAL('triggered()'), 
                             self.__projectBrowserPylintShow)
            
            if not self.__projectBrowserAct in menu.actions():
                menu.addAction(self.__projectBrowserAct)
            if not self.__projectBrowserShowAct in menu.actions():
                menu.addAction(self.__projectBrowserShowAct)
            self.__projectBrowserShowAct.setEnabled(self.__pylintPsbDialog is not None)
    
    def __pyLint(self, project, mpName, forProject, forEditor = False):
        """
        Private method used to perform a PyLint run.
        
        @param project reference to the Project object
        @param mpName name of module or package to be checked (string or QString)
        @param forProject flag indicating a run for the project (boolean)
        """
        if forEditor:
            parms = copy.deepcopy(self.__editorParms)
        else:
            parms = project.getData('CHECKERSPARMS', "PYLINT")
        dlg = PyLintConfigDialog(project.getProjectPath(), parms)
        if dlg.exec_() == QDialog.Accepted:
            args, parms = dlg.generateParameters()
            self.__editorParms = copy.deepcopy(parms)
            if not forEditor:
                project.setData('CHECKERSPARMS', "PYLINT", parms)
            
            # now do the call
            dlg2 = PyLintExecDialog()
            try:
                reportFile = parms['reportFile']
            except KeyError:
                reportFile = None
            res = dlg2.start(args, mpName, reportFile, project.ppath)
            if res:
                dlg2.show()
            if forProject:
                self.__pylintPDialog = dlg2
            elif forEditor:
                self.__editorPylintDialog = dlg2
            else:
                self.__pylintPsbDialog = dlg2
    
    def __projectPylint(self):
        """
        Public slot used to check the project files with Pylint.
        """
        project = e4App().getObject("Project")
        project.saveAllScripts()
        self.__pyLint(project, project.getProjectPath(), True)
    
    def __projectPylintShow(self):
        """
        Public slot to show the PyLint dialog with the results of the last run.
        """
        if self.__pylintPDialog is not None:
            self.__pylintPDialog.show()
    
    def __projectBrowserPylint(self):
        """
        Private method to handle the Pylint context menu action of the project
        sources browser.
        """
        project = e4App().getObject("Project")
        browser = e4App().getObject("ProjectBrowser").getProjectBrowser("sources")
        itm = browser.model().item(browser.currentIndex())
        try:
            fn = itm.fileName()
        except AttributeError:
            fn = itm.dirName()
        self.__pyLint(project, fn, False)
    
    def __projectBrowserPylintShow(self):
        """
        Public slot to show the PyLint dialog with the results of the last run.
        """
        if self.__pylintPsbDialog is not None:
            self.__pylintPsbDialog.show()
    
    def __editorOpened(self, editor):
        """
        Private slot called, when a new editor was opened.
        
        @param editor reference to the new editor (QScintilla.Editor)
        """
        menu = editor.getMenu("Checks")
        if menu is not None:
            menu.addAction(self.__editorAct)
            self.connect(editor, SIGNAL("showMenu"), self.__editorShowMenu)
            self.__editors.append(editor)
    
    def __editorClosed(self, editor):
        """
        Private slot called, when an editor was closed.
        
        @param editor reference to the editor (QScintilla.Editor)
        """
        try:
            self.__editors.remove(editor)
        except ValueError:
            pass
    
    def __editorShowMenu(self, menuName, menu, editor):
        """
        Private slot called, when the the editor context menu or a submenu is 
        about to be shown.
        
        @param menuName name of the menu to be shown (string)
        @param menu reference to the menu (QMenu)
        @param editor reference to the editor
        """
        if menuName == "Checks":
            if not self.__editorAct in menu.actions():
                menu.addAction(self.__editorAct)
            self.__editorAct.setEnabled(editor.isPyFile())
    
    def __editorPylint(self, ):
        """
        Private slot to handle the Pylint context menu action of the editors.
        """
        editor = e4App().getObject("ViewManager").activeWindow()
        if editor is not None:
            if not editor.checkDirty():
                return
        
        fn = editor.getFileName()
        project = e4App().getObject("Project")
        self.__pyLint(project, fn, False, True)
##    
##    def __projectBrowserPylintShow(self):
##        """
##        Public slot to show the PyLint dialog with the results of the last run.
##        """
##        if self.__editorPylintDialog is not None:
##            self.__editorPylintDialog.show()
