# -*- coding: utf-8 -*-

# Copyright (c) 2008 - 2009 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""
Module implementing the Rope refactoring plugin.
"""

import sys
import os

from PyQt4.QtCore import QObject, SIGNAL, QTranslator, QString, QStringList, QVariant, \
    QTimer
from PyQt4.QtGui import QApplication

from KdeQt.KQApplication import e4App

import Preferences
from Preferences.Shortcuts import readShortcuts

# Start-of-Header
name = "Refactoring Rope Plugin"
author = "Detlev Offenbach <detlev@die-offenbachs.de>"
autoactivate = True
deactivateable = True
version = "1.1.5"
className = "RefactoringRopePlugin"
packageName = "RefactoringRope"
internalPackages = "rope"
shortDescription = "Refactoring using the Rope library."
longDescription = """This plugin implements refactoring functionality using the"""\
    """ Rope library."""
needsRestart = True
# End-Of-Header

error = QString("")

refactoringRopePluginObject = None

def createAutoCompletionPage(configDlg):
    """
    Module function to create the autocompletion configuration page.
    
    @return reference to the configuration page
    """
    global refactoringRopePluginObject
    from RefactoringRope.ConfigurationPage.AutoCompletionRopePage \
        import AutoCompletionRopePage
    page = AutoCompletionRopePage(refactoringRopePluginObject)
    return page
    
def createCallTipsPage(configDlg):
    """
    Module function to create the calltips configuration page.
    
    @return reference to the configuration page
    """
    global refactoringRopePluginObject
    from RefactoringRope.ConfigurationPage.CallTipsRopePage \
        import CallTipsRopePage
    page = CallTipsRopePage(refactoringRopePluginObject)
    return page
    
def getConfigData():
    """
    Module function returning data as required by the configuration dialog.
    
    @return dictionary containing the relevant data
    """
    if e4App().getObject("UserInterface").versionIsNewer('4.1.99', '20080426'):
        return {
            "ropeAutoCompletionPage" : \
                [QApplication.translate("RefactoringRopePlugin", "Rope"), 
                 os.path.join("RefactoringRope", "ConfigurationPage", 
                              "preferences-refactoring.png"),
                 createAutoCompletionPage, "editorAutocompletionPage", None],
            "ropeCallTipsPage" : \
                [QApplication.translate("RefactoringRopePlugin", "Rope"), 
                 os.path.join("RefactoringRope", "ConfigurationPage", 
                              "preferences-refactoring.png"),
                 createCallTipsPage, "editorCalltipsPage", None],
        }
    else:
        return {}

def prepareUninstall():
    """
    Module function to prepare for an uninstallation.
    """
    ropeRefactoring = RefactoringRopePlugin(None)
    ropeRefactoring.prepareUninstall()
    
class RefactoringRopePlugin(QObject):
    """
    Class implementing the Rope refactoring plugin.
    """
    def __init__(self, ui):
        """
        Constructor
        
        @param ui reference to the user interface object (UI.UserInterface)
        """
        QObject.__init__(self, ui)
        self.__ui = ui
        self.__initialize()
        
        self.__refactoringDefaults = {
            "CodeAssistEnabled" : 0, 
            "MaxFixes" : 10, 
            "CodeAssistTimeout" : 100, 
            
            "CodeAssistCalltipsEnabled" : 0, 
            "CalltipsMaxFixes" : 10, 
        }
        
        self.__translator = None
        self.__loadTranslator()
        
        self.__acTimer = QTimer(self)
        self.__acTimer.setSingleShot(True)
        self.__acTimer.setInterval(self.getPreferences("CodeAssistTimeout"))
        self.connect(self.__acTimer, SIGNAL("timeout()"), 
                     self.__codeAssist)
    
    def __initialize(self):
        """
        Private slot to (re)initialize the plugin.
        """
        self.__object = None
        
        self.__mainAct = None
        self.__mainMenu = None
        self.__projectIsOpen = False
        
        self.__editors = []
        
        self.__currentEditor = None
        self.__savedEditorName = None
        self.__oldEditorText = ""
    
    def __checkVersions(self):
        """
        Private function to check that the eric4 and Python versions are ok.
        
        @return flag indicating version is ok (boolean)
        """
        global error
        
        if self.__ui.versionIsNewer('4.0.99', '20080117'):
            error = QString("")
        else:
            error = self.trUtf8("eric4 version is too old, %1, %2 or newer needed.")\
                .arg("4.1.0").arg("snapshot-20080118")
            return False
        
        if sys.version_info < (2, 5, 0):
            error = self.trUtf8("Python version must be 2.5.0 or newer.")
            return False
        
        return True
    
    def activate(self):
        """
        Public method to activate this plugin.
        
        @return tuple of None and activation status (boolean)
        """
        if not self.__checkVersions():
            return None, False
        
        global refactoringRopePluginObject
        refactoringRopePluginObject = self
        
        from RefactoringRope.Refactoring import Refactoring
        
        self.__object = Refactoring(self, self.__ui)
        self.__object.initActions()
        e4App().registerPluginObject("RefactoringRope", self.__object)
        try:
            readShortcuts(pluginName = "RefactoringRope")
        except TypeError:
            # backwards comaytibility, ignore
            pass
        
        self.__mainMenu = self.__object.initMenu()
        extrasAct = self.__ui.getMenuBarAction("extras")
        self.__mainAct = self.__ui.menuBar().insertMenu(extrasAct, self.__mainMenu)
        
        if e4App().getObject("Project").isOpen():
            self.__mainAct.setEnabled(True)
            self.__object.projectOpened()
        else:
            self.__mainAct.setEnabled(False)
        
        self.connect(e4App().getObject("ViewManager"), SIGNAL("editorOpenedEd"), 
                     self.__editorOpened)
        self.connect(e4App().getObject("ViewManager"), SIGNAL("editorClosedEd"), 
                     self.__editorClosed)
        
        self.connect(e4App().getObject("Project"), SIGNAL('projectOpened'),
                     self.__object.projectOpened)
        self.connect(e4App().getObject("Project"), SIGNAL('projectClosed'),
                     self.__object.projectClosed)
        self.connect(e4App().getObject("Project"), SIGNAL('newProject'),
                     self.__object.projectOpened)
        
        self.connect(e4App().getObject("Project"), SIGNAL('projectOpened'),
                     self.__projectOpened)
        self.connect(e4App().getObject("Project"), SIGNAL('projectClosed'),
                     self.__projectClosed)
        self.connect(e4App().getObject("Project"), SIGNAL('newProject'),
                     self.__projectOpened)
        
        if e4App().getObject("Project").isOpen():
            for editor in e4App().getObject("ViewManager").getOpenEditors():
                self.__editorOpened(editor)
        
        return None, True
    
    def deactivate(self):
        """
        Public method to deactivate this plugin.
        """
        e4App().unregisterPluginObject("RefactoringRope")
        
        self.disconnect(e4App().getObject("ViewManager"), SIGNAL("editorOpenedEd"), 
                        self.__editorOpened)
        self.disconnect(e4App().getObject("ViewManager"), SIGNAL("editorClosedEd"), 
                        self.__editorClosed)
        
        self.disconnect(e4App().getObject("Project"), SIGNAL('projectOpened'),
                        self.__projectOpened)
        self.disconnect(e4App().getObject("Project"), SIGNAL('projectClosed'),
                        self.__projectClosed)
        self.disconnect(e4App().getObject("Project"), SIGNAL('newProject'),
                        self.__projectOpened)
        
        self.disconnect(e4App().getObject("Project"), SIGNAL('projectOpened'),
                        self.__object.projectOpened)
        self.disconnect(e4App().getObject("Project"), SIGNAL('projectClosed'),
                        self.__object.projectClosed)
        self.disconnect(e4App().getObject("Project"), SIGNAL('newProject'),
                        self.__object.projectOpened)
        
        self.__ui.menuBar().removeAction(self.__mainAct)
        
        for editor in self.__editors[:]:
            self.__editorClosed(editor)
        
        self.__initialize()
    
    def __loadTranslator(self):
        """
        Private method to load the translation file.
        """
        if self.__ui is not None:
            loc = self.__ui.getLocale()
            if loc and loc != "C":
                locale_dir = \
                    os.path.join(os.path.dirname(__file__), "RefactoringRope", "i18n")
                translation = "rope_%s" % loc
                translator = QTranslator(None)
                loaded = translator.load(translation, locale_dir)
                if loaded:
                    self.__translator = translator
                    e4App().installTranslator(self.__translator)
                else:
                    print "Warning: translation file '%s' could not be loaded." % \
                        translation
                    print "Using default."
    
    def getPreferences(self, key):
        """
        Public method to retrieve the various refactoring settings.
        
        @param key the key of the value to get
        @param prefClass preferences class used as the storage area
        @return the requested refactoring setting
        """
        return Preferences.Prefs.settings.value("RefactoringRope/" + key,
            QVariant(self.__refactoringDefaults[key])).toInt()[0]
    
    def setPreferences(self, key, value):
        """
        Public method to store the various refactoring settings.
        
        @param key the key of the setting to be set (string)
        @param value the value to be set
        @param prefClass preferences class used as the storage area
        """
        Preferences.Prefs.settings.setValue("RefactoringRope/" + key, QVariant(value))
        
        if key in ["CodeAssistEnabled", "CodeAssistCalltipsEnabled"]:
            if value:
                if e4App().getObject("Project").isOpen():
                    for editor in e4App().getObject("ViewManager").getOpenEditors():
                        if editor not in self.__editors:
                            self.__editorOpened(editor)
            else:
                for editor in self.__editors[:]:
                    self.__editorClosed(editor)
        elif key == "CodeAssistTimeout":
            self.__acTimer.setInterval(value)
    
    def prepareUninstall(self):
        """
        Public method to prepare for an uninstallation.
        """
        Preferences.Prefs.settings.remove("RefactoringRope")
    
    def __projectOpened(self):
        """
        Public slot to handle the projectOpened signal.
        """
        self.__mainAct.setEnabled(True)
        self.__projectIsOpen = True
        
        if self.__ui.versionIsNewer('4.1.99', '20080628'):
            for editor in self.__editors:
                if editor.getLanguage() == "Python" and \
                   self.__projectIsOpen:
                    self.__connectEditorSignals(editor)
                    if self.getPreferences("CodeAssistEnabled"):
                        self.__setAutoCompletionHook(editor)
                    if self.getPreferences("CodeAssistCalltipsEnabled"):
                        self.__setCalltipsHook(editor)
    
    def __projectClosed(self):
        """
        Public slot to handle the projectClosed signal.
        """
        self.__mainAct.setEnabled(False)
        self.__projectIsOpen = False
        
        if self.__ui.versionIsNewer('4.1.99', '20080628'):
            for editor in self.__editors:
                self.__disconnectEditorSignals(editor)
                if editor.autoCompletionHook() == self.codeAssist:
                    self.__unsetAutoCompletionHook(editor)
                if editor.callTipHook() == self.codeAssistCallTip:
                    self.__unsetCalltipsHook(editor)
    
    def __editorOpened(self, editor):
        """
        Private slot called, when a new editor was opened.
        
        @param editor reference to the new editor (QScintilla.Editor)
        """
        if self.__ui.versionIsNewer('4.1.99', '20080628'):
            if editor.getLanguage() == "Python" and \
               self.__projectIsOpen:
                self.__connectEditorSignals(editor)
                if self.getPreferences("CodeAssistEnabled"):
                    self.__setAutoCompletionHook(editor)
                if self.getPreferences("CodeAssistCalltipsEnabled"):
                    self.__setCalltipsHook(editor)
            
            self.connect(editor, SIGNAL("languageChanged"), self.__editorLanguageChanged)
            self.__editors.append(editor)
    
    def __editorClosed(self, editor):
        """
        Private slot called, when an editor was closed.
        
        @param editor reference to the editor (QScintilla.Editor)
        """
        if self.__ui.versionIsNewer('4.1.99', '20080628'):
            if editor in self.__editors:
                self.disconnect(editor, SIGNAL("languageChanged"), 
                                self.__editorLanguageChanged)
                self.__disconnectEditorSignals(editor)
                self.__editors.remove(editor)
                if editor.autoCompletionHook() == self.codeAssist:
                    self.__unsetAutoCompletionHook(editor)
                if editor.callTipHook() == self.codeAssistCallTip:
                    self.__unsetCalltipsHook(editor)
    
    def __editorLanguageChanged(self, language):
        """
        Private slot to handle the language change of an editor.
        """
        editor = self.sender()
        if language == "Python" and \
           self.__projectIsOpen:
            self.__connectEditorSignals(editor)
            if self.__ui.versionIsNewer('4.1.99', '20080628'):
                if self.getPreferences("CodeAssistEnabled"):
                    self.__setAutoCompletionHook(editor)
                if self.getPreferences("CodeAssistCalltipsEnabled"):
                    self.__setCalltipsHook(editor)
        else:
            self.__disconnectEditorSignals(editor)
            if self.__ui.versionIsNewer('4.1.99', '20080628'):
                if editor.autoCompletionHook() == self.codeAssist:
                    self.__unsetAutoCompletionHook(editor)
                if editor.callTipHook() == self.codeAssistCallTip:
                    self.__unsetCalltipsHook(editor)
    
    def __connectEditorSignals(self, editor):
        """
        Private method to connect to some signals of an editor.
        
        @param editor reference to the editor (QScintilla.Editor)
        """
        self.connect(editor, SIGNAL("editorAboutToBeSaved"), 
                     self.__editorAboutToBeSaved)
        self.connect(editor, SIGNAL("editorSaved"), 
                     self.__editorSaved)
    
    def __disconnectEditorSignals(self, editor):
        """
        Private method to disconnect to some signals of an editor.
        
        @param editor reference to the editor (QScintilla.Editor)
        """
        self.disconnect(editor, SIGNAL("editorAboutToBeSaved"), 
                        self.__editorAboutToBeSaved)
        self.disconnect(editor, SIGNAL("editorSaved"), 
                        self.__editorSaved)
    
    def __completionListSelected(self, id, txt):
        """
        Private slot to handle the selection from the completion list.
        
        @param id the ID of the user list (should be 1) (integer)
        @param txt the selected text (QString)
        """
        from QScintilla.Editor import EditorAutoCompletionListID
        
        editor = self.sender()
        if id == EditorAutoCompletionListID:
            lst = txt.split(" ")
            if len(lst) > 1:
                txt = lst[0]
            
            if Preferences.getEditor("AutoCompletionReplaceWord"):
                editor.selectCurrentWord()
                editor.removeSelectedText()
                line, col = editor.getCursorPosition()
            else:
                line, col = editor.getCursorPosition()
                wLeft = editor.getWordLeft(line, col)
                if not wLeft.isEmpty():
                    txt = txt.replace(wLeft, '')
            editor.insert(txt)
            editor.setCursorPosition(line, col + txt.length())
    
    def __setAutoCompletionHook(self, editor):
        """
        Private method to set the autocompletion hook.
        
        @param editor reference to the editor (QScintilla.Editor)
        """
        self.connect(editor, SIGNAL('userListActivated(int, const QString)'),
                     self.__completionListSelected)
        editor.setAutoCompletionHook(self.codeAssist)
    
    def __unsetAutoCompletionHook(self, editor):
        """
        Private method to unset the autocompletion hook.
        
        @param editor reference to the editor (QScintilla.Editor)
        """
        editor.unsetAutoCompletionHook()
        self.disconnect(editor, SIGNAL('userListActivated(int, const QString)'),
                        self.__completionListSelected)
    
    def codeAssist(self, editor, context = False):
        """
        Public method to determine the autocompletion proposals.
        
        @param editor reference to the editor object, that called this method
            QScintilla.Editor)
        @param context flag indicating to autocomplete a context (boolean)
        """
        self.__currentEditor = editor
        if self.getPreferences("CodeAssistTimeout"):
            self.__acTimer.stop()
            self.__acTimer.start()
        else:
            self.__codeAssist()
    
    def __codeAssist(self):
        """
        Private slot to show a list with completion proposals.
        """
        from RefactoringRope.CodeAssist import CodeAssist
        from QScintilla.Editor import EditorAutoCompletionListID
        
        if self.__currentEditor is not None:
            if self.__currentEditor.isListActive():
                self.__currentEditor.cancelList()
            ca = CodeAssist(self.__object, self.__currentEditor, self, self)
            completions = ca.getCompletions()
            if len(completions) == 0:
                # try QScintilla autocompletion
                self.__currentEditor.autoCompleteQScintilla()
                pass
            else:
                self.__currentEditor.showUserList(EditorAutoCompletionListID, completions)
    
    def __editorAboutToBeSaved(self, filename):
        """
        Private slot to get the old contents of the named file.
        
        @param filename name of the file about to be saved (string)
        """
        if filename and os.path.exists(filename):
            try:
                f = open(filename, "rb")
                self.__oldEditorText = f.read()
            except IOError:
                self.__oldEditorText = ""
            f.close()
            self.__savedEditorName = filename
        else:
            self.__savedEditorName = ""
            self.__oldEditorText = ""
    
    def __editorSaved(self, filename):
        """
        Private slot to activate SOA.
        
        @param filename name of the file that was saved (string)
        """
        import rope.base.libutils
        
        if filename == self.__savedEditorName and self.__oldEditorText:
            rope.base.libutils.report_change(self.__object.getProject(), 
                self.__savedEditorName, self.__oldEditorText)
        elif self.__savedEditorName == "":
            rope.base.libutils.report_change(self.__object.getProject(), 
                filename, "")
    
    def __setCalltipsHook(self, editor):
        """
        Private method to set the calltip hook.
        
        @param editor reference to the editor (QScintilla.Editor)
        """
        editor.setCallTipHook(self.codeAssistCallTip)
    
    def __unsetCalltipsHook(self, editor):
        """
        Private method to unset the calltip hook.
        
        @param editor reference to the editor (QScintilla.Editor)
        """
        editor.unsetCallTipHook()
    
    def codeAssistCallTip(self, editor, pos, commas):
        """
        Public method to return a list of calltips.
        
        @param editor reference to the editor (QScintilla.Editor)
        @param pos position in the text for the calltip (integer)
        @param commas minimum number of commas contained in the calltip (integer)
        @return list of possible calltips (list of strings)
        """
        from RefactoringRope.CodeAssist import CodeAssist
        ca = CodeAssist(self.__object, editor, self, self)
        cts = ca.getCallTips(pos)
        return cts
