# -*- coding: utf-8 -*-

# Copyright (c) 2008 - 2009 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""
Module implementing the Move Module dialog.
"""

import os

from PyQt4.QtGui import *
from PyQt4.QtCore import *

from KdeQt import KQFileDialog, KQMessageBox
from KdeQt.KQApplication import e4App

from E4Gui.E4Completers import E4DirCompleter

from Ui_MoveModuleDialog import Ui_MoveModuleDialog
from RefactoringDialogBase import RefactoringDialogBase

import Utilities

class MoveModuleDialog(RefactoringDialogBase, Ui_MoveModuleDialog):
    """
    Class implementing the Move Module dialog.
    """
    def __init__(self, refactoring, title, mover, parent = None):
        """
        Constructor
        
        @param refactoring reference to the main refactoring object (Refactoring)
        @param title title of the dialog (string or QString)
        @param mover reference to the mover object (rope.refactor.move.MoveMethod)
        @param parent reference to the parent widget (QWidget)
        """
        RefactoringDialogBase.__init__(self, refactoring, title, parent)
        self.setupUi(self)
        
        self.__destinationCompleter = E4DirCompleter(self.destinationEdit)
        
        self.__mover = mover
        self.__projectRootPath = e4App().getObject("Project").getProjectPath()
        
        self.__okButton = self.buttonBox.button(QDialogButtonBox.Ok)
        self.__okButton.setEnabled(False)
        self.__previewButton = self.buttonBox.addButton(\
            self.trUtf8("Preview"), QDialogButtonBox.ActionRole)
        self.__previewButton.setDefault(True)
    
    @pyqtSignature("QString")
    def on_destinationEdit_textChanged(self, txt):
        """
        Slot called, when the destinationEdit text is changed.
        """
        self.__okButton.setEnabled(not txt.isEmpty())
    
    @pyqtSignature("")
    def on_selectButton_clicked(self):
        """
        Slot called to select the destination via a file selection dialog.
        """
        dest = unicode(self.destinationEdit.text())
        if not dest:
            dest = self.__projectRootPath
        elif not os.path.isabs(dest):
            dest = os.path.join(self.__projectRootPath, dest)
        destination = KQFileDialog.getExistingDirectory(\
            self,
            self.windowTitle(),
            dest,
            QFileDialog.Options(QFileDialog.Option(0)))
        
        if not destination.isEmpty():
            destination = Utilities.toNativeSeparators(destination)
            if not destination.startsWith(self.__projectRootPath):
                KQMessageBox.critical(self,
                    self.windowTitle(),
                    self.trUtf8("""The selected directory must be """
                        """inside the project."""))
                return
            
            if not os.path.exists(os.path.join(unicode(destination), "__init__.py")):
                KQMessageBox.critical(self,
                    self.windowTitle(),
                    self.trUtf8("""The selected directory <b>%1</b> is not a package.""")\
                        .arg(destination))
                return
            
            destination = \
                unicode(destination).replace(self.__projectRootPath + os.sep, "")
            self.destinationEdit.setText(destination)
    
    def __checkDestination(self):
        """
        Private method to check the destination entered.
        
        @return flag indicating a valid entry (boolean)
        """
        destination = os.path.join(self.__projectRootPath, 
            unicode(self.destinationEdit.text()))
        if not os.path.exists(os.path.join(destination, "__init__.py")):
            KQMessageBox.critical(self,
                self.windowTitle(),
                self.trUtf8("""The selected directory <b>%1</b> is not a package.""")\
                    .arg(destination))
            return False
        
        return True
    
    @pyqtSignature("QAbstractButton*")
    def on_buttonBox_clicked(self, button):
        """
        Private slot to act on the button pressed.
        
        @param button reference to the button pressed (QAbstractButton)
        """
        if button == self.__previewButton and self.__checkDestination():
            self.previewChanges()
        elif button == self.__okButton and self.__checkDestination():
            self.applyChanges()
    
    def _calculateChanges(self, handle):
        """
        Protected method to calculate the changes.
        
        @param handle reference to the task handle (rope.base.taskhandle.TaskHandle)
        @return reference to the Changes object (rope.base.change.ChangeSet)
        """
        dest = unicode(self.destinationEdit.text())
        if dest.endswith(os.sep):
            dest = dest[:-1]
        destination = self.__mover.project.get_pycore().find_module(dest)
        try:
            changes = self.__mover.get_changes(destination, task_handle = handle)
            return changes
        except Exception, err:
            self._refactoring.handleRopeError(err, self._title, handle)
            return None
